local MAJOR, MINOR = "LibMindReader", 1

if LibStub == nil then
  error(MAJOR .. ' requires LibStub');
end

local mr, oldminor = LibStub:NewLibrary(MAJOR, MINOR)

if not mr then return end -- No Upgrade needed.

mr.embeds = mr.embeds or {} -- what objects embed this lib

--- embedding and embed handling
local mixins = {
	"Build", "GetTempStorage", "AddFtionTip", "AddFtionTips", "ShowBuild"
}

-- mr:Embed( target )
-- target (object) - target object to embed LibMindReader in
--
-- Embeds LibMindReader into the target object making the functions from the mixins list available on target:..
function mr:Embed(target)
	for k, v in pairs(mixins) do
		if target[v] then
			error('Cannot mixin '..v..'; member already exists');
		end
		target[v] = self[v];
	end
	self.embeds[target] = true
	return target
end

-- mr:OnEmbedDisable( target )
-- target (object) - target object that is being disabled
--
function mr:OnEmbedDisable(target)
	-- No clue what to do here, but other libs seem to have this...
	-- Perhaps I should clean up target's logs
	if temporary[target] then
		temporary[target] = hp:del(temporary[target]);
	end
end

--[[
	LibMindReader functions
--]]

-- Table to hold the built tables in
local temporary = {};

-- table to hold the function tips in
local ftips = {};

--[[
MakeTableNode()
Generates a single node for a table

	key() = the name that will be given to the node
	parentval(table) = the table this variable belongs to, may be nil
	val(table) = the variable itself
	setup(table) = the setup-table where the variables for the controls are kept
	
returns the created node
--]]
local function MakeTableNode(key, parentval, val, setup)

	if key == nil then
		error('key is nil');
	end
	local sname = tostring(val);
	local propersname = string.gsub(sname,' ','');
	local s = {};
	setup[sname] = s;

	return {
		name = tostring(key),
		type = 'group',
		args = {
			[propersname] = { -- inline group that contains controls for doing stuff on the variable
				name = sname,
				type = 'group',
				order = 1,
				inline = true,
				args = {
					label = {
						name = 'Table variable',
						type = 'description',
						order = 1,
					},
					execinput = {
						name = 'Code',
						type = 'input',
						order = 2,
						multiline = 8,
						arg = { s = s },
						get = function(info)
							return info.arg.s.execinput;
						end,
						set = function(info, value)
							info.arg.s.execinput = value;
						end,
					},
					execbutton = {
						name = 'Execute as Method',
						type = 'execute',
						order = 3,
						arg = {
							s = s,
							t = val,
						},
						func = function(info)
							local f = assert(loadstring(info.arg.s.execinput));
							f(info.arg.t);
						end,
					},
				},
			},
		},
	};

end

--[[
MakeNumberNode()
Generates a single node for a number

	key() = the name that will be given to the node
	parentval(table) = the table this variable belongs to, may be nil
	val(number) = the variable itself
	setup(table) = the setup-table where the variables for the controls are kept
	
returns the created node
--]]
local function MakeNumberNode(key, parentval, val, setup)

	local sname = tostring(parentval)..tostring(key);
	local propersname = string.gsub(sname,' ','');

	local n = {
		name = tostring(key),
		type = 'group',
		args = {
			[propersname] = { -- inline group that contains controls for doing stuff on the variable
				name = sname,
				type = 'group',
				order = 1,
				inline = true,
				args = {
					label = {
						name = 'Number variable',
						type = 'description',
						order = 1,
					},
					numberinput = {
						name = 'Number',
						type = 'input',
						order = 2,
					},
				},
			},
		},
	};
	local numberinput = n.args[propersname].args.numberinput;
	if parentval then
		numberinput.arg = {
			p = parentval,
			k = key,
		};
		numberinput.get = function(info)
			return tostring(info.arg.p[info.arg.k]);
		end;
		numberinput.validate = function(info, value)
			local success, error = pcall(tonumber, value);
			if not success then
				return error;
			end
		end;
		numberinput.set = function(info, value)
			info.arg.p[info.arg.k] = tonumber(value);
		end;
	else
		numberinput.arg = val;
		numberinput.get = function(info)
			return info.arg;
		end;
		numberinput.disabled = true;
	end
	
	return n;

end

--[[
MakeStringNode()
Generates a single node for a string

	key() = the name that will be given to the node
	parentval(table) = the table this variable belongs to, may be nil
	val(string) = the variable itself
	setup(table) = the setup-table where the variables for the controls are kept
	
returns the created node
--]]
local function MakeStringNode(key, parentval, setup)

	local sname = tostring(parentval)..tostring(key);
	local propersname = string.gsub(sname,' ','');

	local n = {
		name = tostring(key),
		type = 'group',
		args = {
			[propersname] = { -- inline group that contains controls for doing stuff on the variable
				name = sname,
				type = 'group',
				order = 1,
				inline = true,
				args = {
					label = {
						name = 'String variable',
						type = 'description',
						order = 1,
					},
					stringinput = {
						name = 'String',
						type = 'input',
						order = 2,
						multiline = 8, -- a string can contain multiple lines
					},
				},
			},
		},
	};
	
	local stringinput = n.args[propersname].args.stringinput;
	if parentval then
		stringinput.arg = {
			p = parentval,
			k = key,
		};
		stringinput.get = function(info)
			return info.arg.p[info.arg.k];
		end;
		stringinput.set = function(info, value)
			info.arg.p[info.arg.k] = value;
		end;
	else
		stringinput.arg = val;
		stringinput.get = function(info)
			return info.arg;
		end;
		stringinput.disabled = true;
	end
	
	return n;

end

--[[
MakeFunctionNode()
Generates a single node for a function
Uses functiontips for having more info on the paramaters to provide

	key() = the name that will be given to the node
	parentval(table) = the table this variable belongs to, may be nil
	val(function) = the variable itself
	setup(table) = the setup-table where the variables for the controls are kept
	path(string) = the path to this function
	
returns the created node
--]]
local function MakeFunctionNode(key, parentval, val, setup, path)

	-- dilemma: do we keep one global setup per function (as functions tend to duplicate rapidly) or one setup per link to a function
	-- I went for option no. 2
	local sname = tostring(parentval)..tostring(key);
	local propersname = string.gsub(sname,' ','');
	-- just in case I change my mind, take into account that the setup may already exist
	if setup[sname] == nil then
		setup[sname] = {};
	end
	local s = setup[sname];
	
	-- Build the parameter-function-creator
	local params = '';
	local paramstring = 'nil';
	if parentval then -- we can pass the parent as self when calling the function
		params = 'self';
	end
	if ftips[val] then
		paramstring = ftips[val];
	end
	s.execinput = [[ return function(]]..params..[[)
   return ]]..paramstring..[[
 end]];

	local n = {
		name = tostring(key),
		type = 'group',
		args = {
			[propersname] = { -- The table has an inline group that contains controls for doing stuff on the variable
				name = sname,
				type = 'group',
				order = 1,
				inline = true,
				args = {
					label = {
						name = 'Function',
						type = 'description',
						order = 1,
					},
					execinput = {
						type = 'input',
						order = 2,
						multiline = 8,
						arg = {	s = s },
						width = 'full',
						get = function(info)
							return info.arg.s.execinput;
						end,
						set = function(info, value)
							info.arg.s.execinput = value;
						end,
					},
					execbutton = {
						name = 'Execute',
						type = 'execute',
						order = 3,
					},
					script = {
						name = 'Script',
						type = 'input',
						width = 'full',
						arg = '/script '..path..'('..params..')',
						get = function(info)
							return info.arg
						end,
					},
				},
			},
		},
	};
	
	local execinput = n.args[propersname].args.execinput;
	if parentval then
		execinput.name = 'Code (first param = self)';
	else
		execinput.name = 'Code (no first param given)';
	end
	
	local execbutton = n.args[propersname].args.execbutton;
	if parentval then
		execbutton.arg = {
			s = s,
			f = val,
			p = parentval,
		};
		execbutton.func = function(info)
			-- parse the parameter-function-creator
			local pfc = assert(loadstring(info.arg.s.execinput));
			-- Create the parameter-function
			local pf = pfc();
			-- pass the parenttable as the first argument to it to get the arguments
			-- this way the function can return it as the first parameter and call the ftion as a method
			local p = pf(info.arg.p);
			-- execute the ftion
			info.arg.f(p);
		end
	else
		execbutton.arg = {
			s = s,
			f = val,
		};
		execbutton.func = function(info)
			-- parse the parameter-function-creator
			local pfc = assert(loadstring(info.arg.s.execinput));
			-- Create the parameter-function
			local pf = pfc();
			-- No parenttable, so the function will have to provide the 'self' by itself
			local p = pf();
			-- execute the ftion
			info.arg.f(p);
		end
	end
	
	return n;

end

--[[
:Build()
Build an AceConfigDialog3 optiontable from a variable

	var(object) = the variable
	name(string) = the name of the variable
	
returns the name under which the optiontable will be stored, the table itself and the setup-table where the config stores it's working variables
--]]
function mr:Build(var, name)

	if var == nil then
		return;
	end
	
	if temporary[self] == nil then
		temporary[self] = {};
	end
	
	if name == nil then
		name = tostring(var);
	end
	-- Make a unique name
	local rootname = name .. GetTime();
	
	local t = {
		name = rootname,
		type = 'group',
		args = {},
	};
	
	local s = {}; -- Logs all the vars already encountered, for avoiding up- and cross-links and for keeping workvars for the GUI
	local i = 100; -- Start at 100, this way the children are always AFTER the controls
	local nexti = i + 1; -- Holds the next free spot in the queue
	
	-- To avoid cross-links we work the tree level by level, always adding it's children to the end of the queue
	local q = {}; -- hold the children still left to do, should hold everything needed to process
	q[i] = {
		key = name,
		val = var,
		args = t.args,
		parentval = nil,
		path = name,
	};
	local c, n, pathk;
	
	while q[i] do
		c = q[i]; -- current todo
		if type(c.val) == 'table' then
			-- Check if we haven't seen this table yet
			if not s[tostring(c.val)] then
				-- n holds the new node
				n = MakeTableNode(c.key, c.parentval, c.val, s);
				n.order = i;
				c.args[string.gsub(c.key,' ','')] = n;
				
				-- Add the current var's children to the todo
				for k, v in pairs(c.val) do
					if type(k) == 'string' then
						pathk = "['"..k.."']";
					else
						pathk = '['..tostring(k)..']';
					end
					q[nexti] = {
						key = k,
						val = v,
						args = n.args,
						parentval = c.val,
						path = c.path..pathk,
					};
					nexti = nexti + 1;
				end
			end
		
		elseif type(c.val) == 'number' then
			n = MakeNumberNode(c.key, c.parentval, c.val, s);
			n.order = i;
			c.args[string.gsub(c.key,' ','')] = n;
		
		elseif type(c.val) == 'string' then
			n = MakeStringNode(c.key, c.parentval, c.val, s);
			n.order = i;
			c.args[string.gsub(c.key,' ','')] = n;
		
		elseif type(c.val) == 'function' then
			n = MakeFunctionNode(c.key, c.parentval, c.val, s, c.path);
			n.order = i;
			c.args[string.gsub(c.key,' ','')] = n;
		
		end
		
		-- Clean up nicely
		q[i] = nil;
		i = i + 1;
	end
	
	local nt = {
		table = t,
		setup = s,
	};
	temporary[self][rootname] = nt;
	
	return rootname, t, s;

end

--[[
:GetTempStorage()

returns the table where the Build-results are stored in for the current addon
]]--
function mr:GetTempStorage()

	if temporary[self] == nil then
		temporary[self] = {};
	end
	
	return temporary[self];

end

--[[
:GetAllStorage()

return the whole table where the Build-results are stored, this isn't mixed in
]]--
function mr:GetAllStorage()

	return temporary;

end

--[[
:AddFtionTip()

	func(function) = the function in question
	paramstring(string) = the parameterstring of the function
	ismethod(boolean) = indicated if there is an extra 'self' parameter to be added
	
Adds the function tip to the table
--]]
function mr:AddFtionTip(func, paramstring, ismethod)
	if ismethod then
		paramstring = 'self, '..paramstring;
	end
	ftips[func] = paramstring;
end

--[[
:AddFtionTips()

	tiptable(table) = table with a whole lotta tips
	
Adds the tips to our own tip-table
]]--
function mr:AddFtionTips(tiptable)

	for func, paramstring in pairs(tiptable) do
		ftips[func] = paramstring;
	end

end

--[[
:ShowBuild()
IF AceConfigDialog3 is found this shows the provided table as a GUI. If not, it does nothing.

	rootname(string) = name of Build, may be nil, if not nil and the table is found in the storage, tab is ignored
	tab(table) = the GUI-table to show, if name is nil, this needs to have a value
	
]]--
function mr:ShowBuild(rootname, tab)

	local acd = LibStub("AceConfigDialog-3.0");
	local ac = LibStub("AceConfig-3.0");
	if acd then
		if rootname and temporary[self] and temporary[self][rootname] then
			tab = temporary[self][rootname];
		end
		if tab == nil then
			if rootname then
				error('No table found to show for '..rootname);
			else
				error('No table found to show');
			end
		end
		if rootname == nil then
			rootname = tostring(tab);
		end
		
		ac:RegisterOptionsTable(rootname, tab.table);
		acd:Open(rootname);
	end

end


--[[
	Rest of the lib-stuff
--]]

-- what's this for anyway? isn't mr.embeds empty at this time??
for addon in pairs(mr.embeds) do
	mr:Embed(addon)
end